import os, json, re, sys, subprocess, platform
from . import update_list

# check what os the library is running on
import os
import platform

def check_os():
    os_name = os.name
    system_name = str(platform.system()).lower()
    if os_name == "nt" or system_name == "darwin":
        pass
    else:
        raise Exception("AppOpener only works on Windows or MacOS.")
        exit()

# Get path of working directory
def get_path():
    if getattr(sys, 'frozen', False):
        main_path = os.path.dirname(sys.executable)
        return main_path
    elif __file__:
        main_path = os.path.dirname(__file__)
        return main_path

main_path = os.path.join(get_path(), "Data")

# Status of existance of required files
check_data = os.path.isdir(main_path)
check_json_list = os.path.exists(os.path.join(main_path,"data.json"))
check_app_names = os.path.exists(os.path.join(main_path,"app_names.json"))

# Convert text file to json format file
def setup_files():
    with open((os.path.join(main_path,"data.json")),"r") as data_file:
        data_duplicate = json.load(data_file)
        modified_data = {}
        for key, value in data_duplicate.items():
            is_digit = key[:1].isdigit()
            if is_digit == False:
                val=(re.compile(r'[^a-z-&]')).sub(" ",key)
                final_app_name = re.sub(' +', ' ', val).strip()
                modified_data[final_app_name] = value
            if is_digit == True:
                val=(re.compile(r'[^a-z-&^0-9]')).sub(" ",key)
                final_app_name = re.sub(' +', ' ', val).strip()
                modified_data[final_app_name] = value
        with open((os.path.join(main_path,"data.json")),"w") as f:
            json.dump(modified_data, f, indent=4)

def create_file(print_text=True):
    if check_data == False:
        os.mkdir(main_path)
        subprocess.call("attrib +h "+main_path, shell=True)
    if print_text:
        print("LOADING APPS... (JUST ONCE)")
    cmd = 'powershell -ExecutionPolicy Bypass "Get-StartApps|convertto-json"'
    try:
        result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, encoding='utf-8', shell=True)
        # Exception if the result is Empty
        # if not result.stdout:
        #     raise Exception("No output returned by the command")
        apps = json.loads(result.stdout)
    except:
        from . import old_style
        return
    names = {}
    for each in apps:
        names.update({each['Name'].lower():each['AppID']})
    with open((os.path.join(main_path,"data.json")),"w") as outfile:
        json.dump(names,outfile,indent = 4)
    setup_files()

def check_json():
    if check_json_list == (False):
        create_file()

# Make seperate file for appnames (to perform various features)
def app_names():
    if check_app_names == (False):
        update_list.check_app_names()
    update_list.check_new_name()
    update_list.pre_change()
    update_list.modify()



__version__ = "1.8"
from . import check, update_list, commands, features
from difflib import SequenceMatcher
import os, re, subprocess

check.check_os()
# Checking if required files exists or not
if os.name == "nt":
    check.check_json()
    check.app_names()

    # Get the path of working directory
    main_path = os.path.join((check.get_path()), "Data")

    # REDIRECTED FUNCTION FOR SIMPLIFICATION
    def give_appnames(upper=False):
        '''
        #### `give_appname` is a callable function which gives Appnames as Dictionary in upper or lower case.
        Examples:
        1: app_names = give_appnames(upper=True)
        2: app_names = give_appnames(upper=False)
        #### `upper=True` gives appnames in upper case.
        #### `upper=False` gives appnames in lower case.
        '''
        if upper:
            upper = True
        keys = features.give_appnames(upper=upper)
        return keys

    # For making list
    def mklist(name="", path="", output=True):
        '''
        #### `mklist` is the function which gives Appnames along with their Appids in the specified output file.
        Examples:
        1: mklist(name="json_format.json", path="path_to_folder", output=True)
        2: mklist(name="txt_format.txt", path="path_to_folder", output=True)
        #### `path=` is the path where the file must be stored (to save in working dir dont include it)
        #### `output=True` prints the context of funtion.
        '''
        name = name
        path = path
        output = output
        features.mklist(name=name, path=path, output=output)

    # Run application (Regex implemented)
    def run(self, output=True):
        print()
        print("'RUN' FUNCTION IS REPLACED BY 'OPEN' FUNCTION")
        print("TRY USING 'OPEN(app_name)'")
        print()

    # Open application (Regex implemented)
    def open(self, output=True, match_closest=False, throw_error=False):
        '''
        #### `open` is the function which is used to open applications.
        Examples:
        1: open("whatsapp")
        2: open("whatsapp, telegram")
        3: open("calcu", match_closest=True)
        4: open("whatsapp, telegram", output=False)
        #### `match_closest=True` matches the closest match in given appname.
        #### `output=False` don't print the context of function.
        '''
        if not output:
            output = False
        if match_closest:
            match_closest = True
        inp = (self).lower()
        val=(re.compile(r'[^a-zA-Z-^0-9?,>&]').sub(" ",inp)).strip()
        if val == (""):
            pass
        elif val == ("cls"):
            os.system("cls")
        elif val == ("version"):
            print("AppOpener version "+__version__)
        elif inp == ("?"):
            invsys = '"'
            os.system(f"explorer {invsys}https://appopener.readthedocs.io/en/latest/{invsys}")
        elif val == ("help"):
            print()
            commands.commands()
            print()
        elif val == ("ls"):
            features.list_apps()
        elif val == ("rename -m"):
            os.startfile(os.path.join(main_path,"app_names.json"))
            print("RELOAD PROGRAM TO APPEND CHANGES")
        elif val == ("update"):
            update_list.update(output=output)
        elif " > " in val:
            update_list.do_changes_cli(val)
            update_list.check_new_name()
            update_list.pre_change()
            update_list.modify()
        elif val == ("default"):
            update_list.default(output=output)
            update_list.check_new_name()
            update_list.pre_change()
            update_list.modify()
        elif val == "mklist":
            mklist(output=output)
        elif "find " in val:
            print()
            val2 = val.replace("find ","")
            empty_list = []
            if "," in val2:
                splited = val2.split(",")
                empty_list.extend(splited)
            else:
                empty_list.append(val2)
            features.find_apps(app_names=empty_list)
            print()
        elif val == "log -c" or val == "log":
            print()
            val2 = val.replace("log -","")
            features.change_log(val)
            print()
        else:
            if "," in val:
                splited = val.split(",")
                for i in splited:
                    j = i.strip()
                    if j != "":
                        features.open_things(j, output=output, match_closest=match_closest, throw_error=throw_error)
            else:
                features.open_things(val, output=output, match_closest=match_closest, throw_error=throw_error)

    # Close any application by just its name :)
    def close(self, output=True, match_closest=False, throw_error=False):
        '''
        #### `close` is the function which is used to close applications.
        Examples:
        1: close("whatsapp")
        2: close("whatsapp, telegram")
        3: close("calcu", match_closest=True)
        4: close("whatsapp, telegram", output=False)
        #### `match_closest=True` matches the closest match in given appname.
        #### `output=False` don't print the context of function.
        '''
        if not output:
            output = False
        if match_closest:
            match_closest = True
        inp = (self).lower()
        val=(re.compile(r'[^a-zA-Z-^0-9?,>&+.]').sub(" ",inp)).strip()
        if "," in val:
            splited = val.split(",")
            for i in splited:
                j = i.strip()
                if j != "":
                    features.close_things(j, output=output, match_closest=match_closest, throw_error=throw_error)
        else:
            features.close_things(val, output=output, match_closest=match_closest, throw_error=throw_error)

else:
    def open(self):

        app = self.lower()
        
        native_list = ['app store', 'mail', 'contacts', 'calender', 'reminders', 'notes', 'face time', 'messages',
                    'maps', 'find my', 'photos', 'photo booth', 'preview', 'music', 'tv', 'podcasts', 'voice memos', 'garageband']
        # highest_similarity = 0.0
        # most_similar_string = None
        # for string in native_list:
        #     similarity_ratio = SequenceMatcher(None, app, string).ratio()
        #     if similarity_ratio > highest_similarity:
        #         highest_similarity = similarity_ratio
        #         most_similar_string = string
        #     print(most_similar_string)
        if app in native_list:
            subprocess.run(["open", "-a", app])

        else:

            item_list = []
            for item in os.listdir("/Applications/"):
                item_path = os.path.join("/Applications/", item)
                item_list.append(item_path)
                print(item_path)

            highest_similarity = 0.0
            most_similar_string = None

            for string in item_list:
                similarity_ratio = SequenceMatcher(None, app, string).ratio()
                if similarity_ratio > highest_similarity:
                    highest_similarity = similarity_ratio
                    most_similar_string = string
            formatted_file_path = f'"{most_similar_string}"'
            os.system(f'open {formatted_file_path}')


    def close(self):

        app = self.lower()

        native_list = ['app store', 'mail', 'contacts', 'calender', 'reminders', 'notes', 'face time', 'messages',
                    'maps', 'find my', 'photos', 'photo booth', 'preview', 'music', 'tv', 'podcasts', 'voice memos', 'garageband']
        # highest_similarity = 0.0
        # most_similar_string = None
        # for string in native_list:
        #     similarity_ratio = SequenceMatcher(None, app, string).ratio()
        #     if similarity_ratio > highest_similarity:
        #         highest_similarity = similarity_ratio
        #         most_similar_string = string
        #     print(most_similar_string)
        if app in native_list:
            subprocess.run(
                ["osascript", "-e", f'tell application "{app}" to quit'])

        else:

            item_list = []
            for item in os.listdir("/Applications/"):
                item_path = os.path.join("/Applications/", item)
                item_list.append(item_path)
                print(item_path)

            highest_similarity = 0.0
            most_similar_string = None

            for string in item_list:
                similarity_ratio = SequenceMatcher(None, app, string).ratio()
                if similarity_ratio > highest_similarity:
                    highest_similarity = similarity_ratio
                    most_similar_string = string
            os.system(
                f'osascript -e \'tell application "{most_similar_string}" to quit\'')




import os, json, sys, psutil, subprocess, difflib, win32gui, win32con

# Defining a custom error message called AppNotFound
class AppNotFound(Exception):
    def __init__(self, app_name):
        self.app_name = app_name
    
    def __str__(self):
        return f"{self.app_name} is not running"

# COLORSHEET FOR TERMINAL WARNINGS !
class style():
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    UNDERLINE = '\033[4m'
    RESET = '\033[0m'

subprocess.call("", shell=True)

# Get path of working directory
def get_path():
    if getattr(sys, 'frozen', False):
        main_path = os.path.dirname(sys.executable)
        return main_path
    elif __file__:
        main_path = os.path.dirname(__file__)
        return main_path

main_path = os.path.join(get_path(),"Data")

# FUNCTION FOR OPENING APPLICATIONS
def open_things(self, output=True, match_closest=False, throw_error=False):
    with open ((os.path.join(main_path,"data.json")),"r") as f:
        data1 = json.load(f)
        keys = data1.keys()
        try:
            dir01 = data1[self]
            subprocess.call("explorer shell:appsFolder\\"+dir01, shell=True)
            # print(("explorer shell:appsFolder\\"+dir01))
            if output:
                print("OPENING "+self.upper())
        except:
            result = difflib.get_close_matches(self,keys, n=1, cutoff=0.6)
            app_name = ' '.join(result).strip()
            if result:
                if output:
                    print()
                    print("Closest match to "+self.upper()+" : "+str(result))
                    print(f'Try : [open("{self}", match_closest=True)]')
                    print()
                if match_closest:
                    dir01 = data1[app_name]
                    subprocess.call("explorer shell:appsFolder\\"+dir01, shell=True)
                    # print("explorer shell:appsFolder\\"+app_name)
                    if output:
                        print("OPENING "+app_name.upper())
            else:
                if output:
                    if throw_error:
                        # Throwing an error if throw_error is true
                        raise AppNotFound(self.upper())
                    else:
                        print(f"{self.upper()} NOT FOUND... TYPE (LS) for list of applications.")
            pass

# EXCEPTION FOR CLOSING FILE EXPLORER
def close_explorer(output: bool, throw_error: bool, app_name: str):
    handles = []
    win32gui.EnumWindows(lambda hwnd, param: param.append(hwnd) if win32gui.GetClassName(hwnd) == 'CabinetWClass' else None, handles)
    # close all File Explorer windows
    for handle in handles:
        win32gui.PostMessage(handle, win32con.WM_CLOSE, 0, 0)
    if not handles:
        if output:
            print((app_name.replace(".exe","")).upper() +" is not running")
    else:
        if output:
            print("CLOSING "+(app_name.replace(".exe","")).upper())
    if throw_error:
        app_name = (app_name).upper()
        raise AppNotFound(app_name)

# CLOSE SEVERAL THINGS :)
def close_things(self, output=True, match_closest=False, throw_error=False):
    if not self.endswith(".exe"):
        self = (self+".exe")
    explorer_list = ["file explorer", "explorer"]
    flag = False
    if not match_closest:
        if (self.replace(".exe", "").strip()) in explorer_list:
            close_explorer(output=output, throw_error=throw_error, app_name=(self.replace(".exe","")))
            return
        for pid in psutil.pids():
            try:
                process = psutil.Process(pid)
                if process.name().lower() == self:
                    process.kill()
                    if not flag and output:
                        print("CLOSING "+(self.replace(".exe","")).upper())
                        flag = True
            except: pass
        if not flag and output:
            if throw_error:
                app_name = (self.replace(".exe","")).upper()
                raise AppNotFound(app_name)
            else:
                print((self.replace(".exe","")).upper() +" is not running")
    if match_closest:
        app_jug = []
        for pid in psutil.pids():
            try:
                process = psutil.Process(pid)
                app_jug.append((process.name()))
            except: pass
        result = difflib.get_close_matches(self,app_jug, n=1, cutoff=0.6)
        app_name = ' '.join(result).strip()
        # print(app_jug)
        # print(app_name)
        ## EXCEPTION FOR CLOSING FILE EXPLORER
        if (app_name.replace(".exe", "").strip()) in explorer_list:
            # print("yes")
            close_explorer(output=output, throw_error=throw_error, app_name=app_name)
            return
        command = ['taskkill', '/f', '/im',app_name]
        # print(command)
        with open('NUL', 'w') as null:
            process = subprocess.Popen(command, stdout=null, stderr=null)
            process.wait()
            if process.returncode == 0:
                if output:
                    print("CLOSING "+(app_name.replace(".exe","")).upper())
            else:
                if output:
                    if throw_error:
                        app_name = (self.replace(".exe", "")).upper()
                        raise AppNotFound(app_name)
                    else:
                        print((self.replace(".exe","")).upper() +" is not running")

# FIND APPLICATION IF INSTALLED OR NOT :)
def find_apps(app_names: list):
    with open((os.path.join(main_path,"app_names.json")),"r") as f:
        data = json.load(f)
        keys = data.keys()
        empty_list = []
        for i in app_names:
            result = difflib.get_close_matches(i,keys, cutoff=0.5)
            empty_list.extend(result)
        num = 0
        if not empty_list:
            print("NOTHIING LIKE THAT INSTALLED IN PC")
            return
        print("FOUND THESE APPLICATIONS INSTALLED: ")
        print()
        for app_name in empty_list:
            num += 1
            print(str(num)+". "+app_name.upper())

# SEE PETANME(s) OF ORIGINAL APP(s)
def change_log(self):
    subprocess.call("", shell=True)
    if self == "log":
        with open((os.path.join(main_path,"app_names.json")),"r") as f:
            data = json.load(f)
            keys = list(data.keys())
            values = list(data.values())
            for i in values:
                if i != "":
                    position = values.index(i)
                    app = keys[position]
                    print(style.RED+(app.upper())+style.WHITE+" > "+style.GREEN+(i.upper())+style.RESET)
    else:
        with open((os.path.join(main_path,"app_names_temp.json")),"r") as f:
            data = json.load(f)
            keys = list(data.keys())
            values = list(data.values())
            for i in values:
                if i != "":
                    position = values.index(i)
                    app = keys[position]
                    print(style.RED+(app.upper())+style.WHITE+" > "+style.GREEN+(i.upper())+style.RESET)

# LISTING APP(s) LIST - 2
def list_apps():
    print()
    with open((os.path.join(main_path,"data.json")),"r") as file:
        data = json.load(file)
        key = data.keys()
        keys = sorted(key)
        count = 0
        for app in keys:
            if len(app.strip()) == 0 :
                continue
            else:
                count += 1
                print("{}. {}".format(count, app.strip().upper()))
    print()

# Give dictionary of appnames (Uppercase or lowercase)
def give_appnames(upper=False):
    with open((os.path.join(main_path,"data.json")),"r") as file:
        data = json.load(file)
    keys = data.keys()
    if upper == True:
        dict = {}
        for k in keys:
            change = {k.upper() : None}
            dict.update(change)
        keys_upper = dict.keys()
        return keys_upper
    if upper == False:
        return keys

# Function for making list
def mklist(name="", path="", output=True):
    path_exists = os.path.isdir(path)
    flag = False
    if not path or not path_exists:
        cwd = os.getcwd()
        path = cwd
        flag = True
    if not name:
        name = "app_data.json"
    if name.endswith(".txt"):
        name = name
    dictionary ={}
    with open((os.path.join(path,name)),"w") as outfile:
        json.dump(dictionary,outfile,indent = 4)
    with open((os.path.join(main_path,"data.json")),"r") as data_f:
        data = json.load(data_f)
    with open((os.path.join(path,name)),"a+") as f:
        g = open((os.path.join(path,name)),"r+")
        g.truncate(0)
        json.dump(data,f,indent=4)
    if output:
        if flag:
            print("Successfully saved "+name+" in the working directory.")
        if not flag:
            print("Successfully saved "+name+" in the given directory.")



## THIS METHOD IS EXECUTED WHEN MODULE THROWS ERROR WHILE CREATING FILES

import os, sys, json, win32gui, win32con, re

# Get path of working directory
def get_path():
    if getattr(sys, 'frozen', False):
        main_path = os.path.dirname(sys.executable)
        return main_path
    elif __file__:
        main_path = os.path.dirname(__file__)
        return main_path

main_path = os.path.join(get_path(), "Data")

check_reference_file = os.path.exists(os.path.join(main_path,"reference.txt"))
check_json_list = os.path.exists(os.path.join(main_path,"data.json"))

def maximize():
    hwnd = win32gui.GetForegroundWindow()
    win32gui.ShowWindow(hwnd, win32con.SW_MAXIMIZE)

def convert_txt_json():
    dictionary ={}
    with open((os.path.join(main_path,"data.json")),"w") as outfile:
        json.dump(dictionary,outfile,indent = 4)
    file1 = open(os.path.join(main_path,'reference.txt'),'r')
    Lines = file1.readlines()
    for line in Lines:
        line1= line.strip()
        index = line1.find('  ')
        # HERE line1[:index] is the APP-NAME
        # HERE line1[index:] is the APP-ID.
        app_name = line1[:index]
        app_id = (line1[index:]).strip()
        is_digit = app_name[:1].isdigit()
        if app_name != "----" and app_name != "":
            with open ((os.path.join(main_path,"data.json")),"r") as f:
                    data = json.load(f)
            if is_digit == (False):
                val=(re.compile(r'[^a-z-&]')).sub(" ",(app_name.lower()))
                final_app_name = re.sub(' +', ' ', val).strip()
                change = {final_app_name:app_id}
                data.update(change)
                with open((os.path.join(main_path,"data.json")),"a+") as f:
                    g = open((os.path.join(main_path,"data.json")),"r+")
                    g.truncate(0)
                    json.dump(data,f,indent=4)
            elif is_digit == (True):
                val=(re.compile(r'[^a-z-&^0-9]')).sub(" ",(app_name.lower()))
                final_app_name = re.sub(' +', ' ', val).strip()
                change = {final_app_name:app_id}
                data.update(change)
                with open((os.path.join(main_path,"data.json")),"a+") as f:
                    g = open((os.path.join(main_path,"data.json")),"r+")
                    g.truncate(0)
                    json.dump(data,f,indent=4)

def check_json():
    if check_json_list == (False):
        print("REARRANGING APPS... (JUST ONCE)")
        convert_txt_json()
        try: os.remove(os.path.join(main_path,"reference_temp.txt"))
        except: pass
    
def check_reference():                               
    if check_reference_file == (False):
        os.system(f"mkdir {main_path}")
        maximize()
        os.system("mode 800")
        print("LOADING APPS... (JUST ONCE)")
        os.system("powershell -command "+'"'+"get-StartApps | Out-File -encoding ASCII -Filepath "+"'"+(os.path.join(main_path,"reference.txt"))+"'"+'"')
        with open((os.path.join(main_path,"reference.txt")),"r") as f, open((os.path.join(main_path,"reference_temp.txt")),"w+") as outfile:
            for i in f.readlines():
                if not i.strip():
                    continue
                if i:
                    outfile.write(i)
        os.system("mode 100")

check_reference()
check_json()